'''
Modified from https://raw.githubusercontent.com/pytorch/vision/v0.9.1/torchvision/models/resnet.py
BSD 3-Clause License
Copyright (c) Soumith Chintala 2016,
All rights reserved.
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
import sys
import torch
import torch.nn as nn
try:
    from torch.hub import load_state_dict_from_url
except ImportError:
    from torch.utils.model_zoo import load_url as load_state_dict_from_url

from functools import partial
from typing import Dict, Type, Any, Callable, Union, List, Optional

negative_a = 0.0
positive_a = 0.001


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(BasicBlock, self).__init__()
        self.inplanes = inplanes
        self.planes = planes
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.S1 = nn.Linear(inplanes * 9 * planes, planes)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.S2 = nn.Linear(planes * 9 * planes, planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, ins):
        x = ins[0]
        scores = ins[1]
        is_score_training = ins[2]

        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        scores11 = torch.exp(torch.flatten(self.S1(torch.flatten(self.conv1.weight))))
        scores11 = torch.where(scores11 < 1, scores11, 1.0 + (scores11 - 1.0) * positive_a)

        if is_score_training == 1:
            scores11 = torch.where(scores11 > 0.5, 1.0 + (scores11) * negative_a, scores11 * negative_a)
        elif is_score_training == 2:
            scores11 = torch.where(scores11 > 0.0, 1.0 + (scores11) * negative_a, scores11 * negative_a)

        out = torch.mul(
            torch.reshape(scores11.repeat(out.size()[0] * out.size()[-1] * out.size()[-2], 1), out.permute(0, 2, 3, 1).size()),
            out.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        scores21 = torch.exp(torch.flatten(self.S2(torch.flatten(self.conv2.weight))))
        scores21 = torch.where(scores21 < 1, scores21, 1.0 + (scores21 - 1.0) * positive_a)

        if is_score_training == 1:
            scores21 = torch.where(scores21 > 0.5, 1.0 + (scores21) * negative_a, scores21 * negative_a)
        elif is_score_training == 2:
            scores21 = torch.where(scores21 > 0.0, 1.0 + (scores21) * negative_a, scores21 * negative_a)

        out = torch.mul(
            torch.reshape(scores21.repeat(out.size()[0] * out.size()[-1] * out.size()[-2], 1),
                          out.permute(0, 2, 3, 1).size()),
            out.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)



        scores.append(scores11)
        scores.append(scores21)

        return (out, scores, is_score_training)


class CifarResNet(nn.Module):

    def __init__(self, block, layers, num_classes=10):
        super(CifarResNet, self).__init__()
        self.inplanes = 16
        self.conv1 = conv3x3(3, 16)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)
        self.S1 = nn.Linear(3 * 9 * 16, 16)

        self.layer1 = self._make_layer(block, 16, layers[0])
        self.layer2 = self._make_layer(block, 32, layers[1], stride=2)
        self.layer3 = self._make_layer(block, 64, layers[2], stride=2)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(64 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x, is_score_training):
        scores = []
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        scores11 = torch.exp(torch.flatten(self.S1(torch.flatten(self.conv1.weight))))
        scores11 = torch.where(scores11 < 1, scores11, 1.0 + (scores11 - 1.0) * positive_a)

        if is_score_training == 1:
            scores11 = torch.where(scores11 > 0.5, 1.0 + (scores11) * negative_a, scores11 * negative_a)
        elif is_score_training == 2:
            scores11 = torch.where(scores11 > 0.0, 1.0 + (scores11) * negative_a, scores11 * negative_a)
        x = torch.mul(
            torch.reshape(scores11.repeat(x.size()[0] * x.size()[-1] * x.size()[-2], 1),
                          x.permute(0, 2, 3, 1).size()),
            x.permute(0, 2, 3, 1)).permute(0, 3, 1, 2)

        scores.append(scores11)

        outs = self.layer1((x, scores, is_score_training))
        outs = self.layer2(outs)
        outs = self.layer3(outs)

        x = outs[0]
        scores = outs[1]
        is_score_training = outs[2]

        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)

        return scores, x


def _resnet(
    arch: str,
    layers: List[int],
    model_urls: Dict[str, str],
    progress: bool = True,
    pretrained: bool = False,
    **kwargs: Any
) -> CifarResNet:
    model = CifarResNet(BasicBlock, layers, **kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls[arch],
                                              progress=progress)
        model.load_state_dict(state_dict)
    return model







